/*

'FLASH' 'S'tandard mode, designed for 'H'itech c (FLASH_SH)

this code accesses flash directly, for testing purposes

Processor:				pic17c43
Compiler:					hi-tech C v7.83pl3
Environment:			compiled with MPLab 4.00.14
Description:			flash routines to communicate with AM29F400 flash
Project setup:		must be in extended microprocessor mode, using ram supplied from target board
									linker options: -L-Pwfh=01FFEh -L-Pesh=01FFCh -L-Pfdih=01FFAh -L-Pefh=01FF8h
Optimisation:			this code is not optimised for speed, it needs some routines converted to
									#defines and the total drop path of procedures shortened
Author:						Shane Tolmie
Tested:						as part of system to download and run code from flash for remote web updates
									data acquisition portion tested OK
									
*/

//put the address of bootrom_write_flash() at the end of the eeprom space
//since a pic17c43 has 4k of rom, this is word 0xFFF, byte 0x1FFE
//or location 0x1FFE in the .hex file
//this must be accompanied by the option -L-Pwfh=01FFEh in the linker
asm("psect wfh,delta=1,size=2");
asm("bootrom_write_flash_hook_label DW _bootrom_write_flash");
#define bootrom_write_flash_hook 0x1FFE/2

//similarly, put the address of bootrom_erase_sector() at the end of the eeprom space
//this must be accompanied by the option -L-Pesh=01FFCh in the linker
asm("psect esh,delta=1,size=2");
asm("bootrom_erase_sector_hook_label DW _bootrom_erase_sector");
#define bootrom_erase_sector_hook 0x1FFC/2

//similarly, put the address of bootrom_flash_device_id() at the end of the eeprom space
//this must be accompanied by the option -L-Pfdih=01FFAh in the linker
asm("psect fdih,delta=1,size=2");
asm("flash_device_id_hook_label DW _bootrom_flash_device_id");
#define bootrom_flash_device_id_hook 0x1FFA/2		//rom is in words, .hex file in bytes

//similarly, put the address of API_EraseFlash() at the end of the eeprom space
//this must be accompanied by the option -L-Pefh=01FF8h in the linker
asm("psect efh,delta=1,size=2");
asm("erase_flash_hook DW _API_EraseFlash");
#define API_EraseFlash_Hook 0x1FF8/2

#include	<pic.h>
#include	"flash_sh.h"
#include	"codes.h"
#include	"always.h"
#include	"delay.h"
#include	"serial.h"

//variables for communication between the program running in flash, and that running in eeprom
bank1 unsigned int	api_flash_uint_control			@ 0x1FE;
bank1 unsigned int	api_flash_uint_address			@	0x1FC;
bank1 unsigned int	api_flash_uint_data					@	0x1FA;
bank1 unsigned char	api_flash_uint_hi_data			@ (0x1FA+1);
bank1 unsigned char	api_flash_uint_lo_data			@ (0x1FA+0);
bank1 unsigned int	api_flash_uint_working 			@ 0x1F8;
bank1 unsigned char	api_flash_uint_hi_working		@ (0x1F8+1);
bank1 unsigned char	api_flash_uint_lo_working		@ (0x1F8+0);
bank1 unsigned char	api_flash_uchar_working 		@ 0x1F7;
bank1 unsigned char	api_flash_uchar_return 			@ 0x1F6;
bank1 unsigned int	api_flash_rwrom_temp 				@ 0x1F4;

void API_SelectFlashBank(unsigned char bank)
{
	FLASH_A16=bank & 0x01;
}

unsigned int write_read_rom_data;

//uses table write to read data from external address lines
//the NOP instructions must be present for reliable operation

//NOTE: this routine has highly picky timing requirements,
//if it doesn't have the exact number of instructions shown
//here, it doesn't work, taking out a nop or adding one 
//makes it cease to perform correctly
//I don't know what will happen if the clock speed is changed!
unsigned int readrom(unsigned int address)
{
	TBLPTRH = (address >> 8);
	TBLPTRL = (address & 0xFF);
	#asm
	MOVLR 0x1																		//switch to bank1
	TABLRD 0,1,_api_flash_rwrom_temp						//dummy read to get it into sync
	TLRD 0,_api_flash_rwrom_temp
	NOP
	TABLRD 1,1,_api_flash_rwrom_temp+1
	NOP
	#endasm
	return api_flash_rwrom_temp;
}

//returns AM29F400 device ID, should be 0x22AB (bottom boot block version)
unsigned int FlashDeviceID(void)
{
	unsigned int device_id;
	
	reset_flash;
	writerom(0xd555,0xaaaa);
	writerom(0xaaaa,0x5555);
	writerom(0xd555,0x9090);
	device_id=readrom(0xFF00);
	reset_flash;
	return device_id;
}

void bootrom_flash_device_id(void)
{
	reset_flash;
	writerom(0xd555,0xaaaa);
	writerom(0xaaaa,0x5555);
	writerom(0xd555,0x9090);
	api_flash_uint_data=readrom(0xFF00);
	reset_flash;
}

unsigned int API_FlashDeviceID(void)
{
	void (*bootrom_flash_device_id_pointer)(void);							//declare pointer to function
	bootrom_flash_device_id_pointer
		=(void (*)(void))ReadFlash(bootrom_flash_device_id_hook);	//hook to vector, see top of this file
	bootrom_flash_device_id_pointer();													//equivalent to API_FlashDeviceID();
	return api_flash_uint_data;
}

//returns manufacturer ID, for AMD it is 0x1
unsigned int API_FlashManufacturerID(void)
{
	unsigned int manu_id;
	
	reset_flash;
	writerom(0xd555,0xaaaa);
	writerom(0xaaaa,0x5555);
	writerom(0xd555,0x9090);
	manu_id=readrom(0xFF01);
	reset_flash;
	return manu_id;
}

//pins A16 and A17 are connected to RA3 and RB5
void API_SetupFlash(void)
{
	unsigned char dummy;

	//read RA3 to set it to an input
	#define FLASH_RY_BYBAR	RA3
	dummy=FLASH_RY_BYBAR;
	
	//select correct bank by driving A16 low
	clrbit(DDRB,5);														//make RB5 output
	FLASH_A16=0;															//set A16 on flash to zero
	reset_flash
}

//erases entire flash, resetting all locations to 0xffff
void API_EraseFlash(void)
{	
	reset_flash;
	writerom(0xd555,0xaaaa);
	writerom(0xaaaa,0x5555);
	writerom(0xd555,0x8080);
	writerom(0xd555,0xaaaa);
	writerom(0xaaaa,0x5555);
	writerom(0xd555,0x1010);
	DelayMs(1);

	while(flash_busy);
	reset_flash;
}

/*

Erases a complete sector in a 64kword bank of code selected by SelectFlashBank(x)

First set the bank using SelectFlashBank(0..3), this sets address lines a16-a17 to get a block of 64kwords
Next, pass any address in the sector required, and it will erase that complete sector
Note that all addresses must be outside the internal rom of the pic, thus to erase sector 0, you must pass
an address in that sector >=0x1000 for the pic17c43, and >0x2000 for the pic17c44

Example of erasing bank 1,

SelectFlashBank(0);
API_EraseSector(0x2000);

Valid banks (anything else will overwrite program code)

Bank		Absolute Address range			SelectFlashBank(x)					API_EraseSector(x)			Safe to use?
0				0x00000 - 0x01FFF						0														can't do it, inside rom	N/A
1				0x02000 - 0x02FFF						0														0x2000									yes
2				0x03000 - 0x03FFF						0														0x3000									yes
3				0x04000 - 0x07FFF						0														0x4000									yes
4				0x08000 - 0x0FFFF						0														0x8000									yes
5				0x10000 - 0x17FFF						1														0x2000									NO, OVERWRITES PROGRAM CODE
6				0x18000 - 0x1FFFF						1														0x8000									yes
7				0x20000 - 0x27FFF						2														0x2000									NO, OVERWRITES PROGRAM CODE
8				0x28000 - 0x2FFFF						2														0x8000									yes
9				0x30000 - 0x37FFF						3														0x2000									NO, OVERWRITES PROGRAM CODE
10			0x38000 - 0x3FFFF						3														0x8000									yes

*/

void EraseSector(unsigned int sector)
{
	reset_flash;
	writerom(0xf555,0xaaaa);
	writerom(0xf2aa,0x5555);
	writerom(0xf555,0x8080);
	writerom(0xf555,0xaaaa);
	writerom(0xf2aa,0x5555);

	writerom(sector,0x3030);
	
	while(flash_busy);
	reset_flash;
	return;
}

void bootrom_erase_sector(void)
{
	reset_flash;
	writerom(0xf555,0xaaaa);
	writerom(0xf2aa,0x5555);
	writerom(0xf555,0x8080);
	writerom(0xf555,0xaaaa);
	writerom(0xf2aa,0x5555);

	writerom(api_flash_uint_data,0x3030);
	
	while(flash_busy);
	reset_flash;
	return;
}

void API_EraseSector(unsigned int sector)
{
	void (*bootrom_erase_sector_pointer)(void);									//declare pointer to function

	api_flash_uint_data=sector;																	//pass sector address at top of ram
	bootrom_erase_sector_pointer
		=(void (*)(void))ReadFlash(bootrom_erase_sector_hook);		//hook to vector, see top of this file
	bootrom_erase_sector_pointer();															//equivalent to EraseSector();
}

//writes to address int addr with int data
unsigned char WriteFlash(unsigned int addr, unsigned int data)
{
	unsigned char counter;
	unsigned char writeflash_uchar_timeout;

	//if routine tries to write to addresses which
	//  are in the PIC17Cx internal memory, the finished
	//  lines will never get toggled, resulting in an endless loop
	if (addr<FLASH_MIN_ADDRESS)
	{
		return FALSE;
	}

	counter=10;
	while(counter-->1)
	{
		//flash requires a preamble before each word programmed
		writerom(0xd555,0xaaaa);
		writerom(0xaaaa,0x5555);
		writerom(0xd555,0xa0a0);

		//put data on address lines
		writerom(addr,data);
		
		//wait for data to be written (usually 15.6us)
		writeflash_uchar_timeout=0;
		#define WRITE_FLASH_TIMEOUT_MAX 0xF0		//@16Mhz, normally 3, F0 gives 1.26ms

		do
		{
			nop();
			nop();
			writeflash_uchar_timeout++;
		} while ( (flash_busy)
				&& (writeflash_uchar_timeout <= WRITE_FLASH_TIMEOUT_MAX) );
		
		if (writeflash_uchar_timeout>=WRITE_FLASH_TIMEOUT_MAX)
		{
			return FALSE;
		}
		#undef WRITE_FLASH_TIMEOUT_MAX

		if(readrom(addr)==data) 												//make sure data was programmed
		{
			return TRUE;																	//if not try to reprogram
		}
	}
	return FALSE;
}

void bootrom_write_flash(void)
{
	//if routine tries to write to addresses which
	//  are in the PIC17Cx internal memory, the finished
	//  lines will never get toggled, resulting in an endless loop

	if (api_flash_uint_address<FLASH_MIN_ADDRESS)
	{
		api_flash_uchar_return=FALSE;
		return;
	}
	
	api_flash_uchar_working=10;
	while(api_flash_uchar_working-->1)
	{
		//flash requires a preamble before each word programmed
		writerom(0xd555,0xaaaa);
		writerom(0xaaaa,0x5555);
		writerom(0xd555,0xa0a0);

		//put data on address lines
		writerom(api_flash_uint_address,api_flash_uint_data);
		
		//wait for data to be written
		while(flash_busy);

		if(readrom(api_flash_uint_address)==api_flash_uint_data)			//make sure data was programmed
		{
			api_flash_uchar_return=TRUE;
			return;																											//if not try to reprogram
		}
	}
	api_flash_uchar_return=FALSE;
	return;
}

unsigned char API_WriteFlash(unsigned int addr, unsigned int data)
{
	void (*bootrom_write_flash_pointer)(void);									//declare pointer to function

	api_flash_uint_address=addr;
	api_flash_uint_data=data;
	bootrom_write_flash_pointer
		=(void (*)(void))ReadFlash(bootrom_write_flash_hook);			//hook to vector, see top of this file
	bootrom_write_flash_pointer();															//equivalent to WriteFlash();

	return api_flash_uchar_return;
}

//reads an array of unsigned ints from ram, &ram_addr to rom, rom_addr, of length length
void API_ReadFlashArray(unsigned int rom_addr, unsigned int *ram_addr, unsigned char length)
{
	unsigned char i;

	for(i=0;i<length;i++)
	{
		ram_addr[i]=API_ReadFlash(rom_addr);
		rom_addr++;
	}
}

//writes an array of unsigned ints from ram, &ram_addr to rom, rom_addr, of length length
unsigned char API_WriteFlashArray(unsigned int rom_addr, unsigned int *ram_addr, unsigned char length)
{
	unsigned char i;

	for(i=0;i<length;i++)
	{
		if (API_WriteFlash(rom_addr,ram_addr[i])==FALSE)
		{
			return FALSE;
		}
		rom_addr++;
	}
	return TRUE;
}

//tests routines so far, outputs lots of 'y' for yes if tests pass
void API_TestRoutines(void)
{
	unsigned int flash_array1[4];
	unsigned int flash_array2[4];
	static unsigned int test;

	#define t(x)	if (test==x) {			\
									putch('y');				\
								}										\
								else {							\
									putch('n');				\
									putch(',');				\
									putinthex(test);	\
									putch('!');				\
									putch('=');				\
									putinthex(x);			\
									putch(' ');				\
								}								
									

	API_SetupFlash();
	API_ResetFlash();
	API_SelectFlashBank(0);						//select first bank of 64kwords out of 256kwords

	test=API_FlashDeviceID();
	b;t(0x0002);											//now test=0x0001, give error on purpose to test it
	
	test=API_FlashManufacturerID();	
	b;t(0x22ab);											//now test=0x22ab

	//test for word/byte boundaries
	API_EraseSector(0x2000);
	API_WriteFlash(0x2000,0x0123);
	API_WriteFlash(0x2001,0x3210);
	test=API_ReadFlash(0x2000);					
	b;t(0x0123);											//now test=0x0123
	test=API_ReadFlash(0x2001);					
	b;t(0x3210);											//now test=0x3210

	//test sector 1, 0x2000 to 0x2fff
	API_EraseSector(0x2000);				
	test=API_ReadFlash(0x2000);					
	b;t(0xffff);											//now test=0xFFFF
	API_WriteFlash(0x2000,0x3456);
	test=API_ReadFlash(0x2000);					
	b;t(0x3456);											//now test=0x3456

	//test sector 2, 0x3000 to 0x3fff

	API_EraseSector(0x3000);				
	test=API_ReadFlash(0x3000);					
	b;t(0xffff);											//now test=0xFFFF
	API_WriteFlash(0x3000,0x1234);
	test=API_ReadFlash(0x3000);					
	b;t(0x1234);											//now test=0x1234

	//test sector 4, 0x8000 to 0xffff
	API_EraseSector(0x8000);				
	test=API_ReadFlash(0x8000);					
	b;t(0xffff);											//now test=0xFFFF
	API_WriteFlash(0x8000,0x2345);
	test=API_ReadFlash(0x8000);					
	b;t(0x2345);											//now test=0x2345

	//make sure that data in sector 2 is still intact
	test=API_ReadFlash(0x3000);					
	b;t(0x1234);											//now test=0x1234


	//test write and read array
	#define OFFSET	0x3000
	API_EraseSector(OFFSET);					//0x3000 to 0x3fff
	test=API_ReadFlash(OFFSET);					
	b;t(0xffff);											//now test=0xFFFF

	flash_array1[0]=0x1231;
	flash_array1[1]=0x1232;
	flash_array1[2]=0x1233;
	flash_array1[3]=0x1234;
	b;
	if (API_WriteFlashArray(OFFSET,&flash_array1[0],4)==FALSE)
	{
		b;															//should never stop here
		putch('n');
	}
	b;
	test=API_ReadFlash(OFFSET);
	test=API_ReadFlash(OFFSET+1);
	b;t(0x1232);											//now test=0x1232
	test=API_ReadFlash(OFFSET+2);
	test=API_ReadFlash(OFFSET+3);	
	b;t(0x1234);											//now test=0x1234

	API_ReadFlashArray(OFFSET,&flash_array2[0],4);
	test=flash_array2[0];
	test=flash_array2[1];
	b;t(0x1232);											//now test=0x1232
	test=flash_array2[2];
	test=flash_array2[3];
	b;t(0x1234);											//now test=0x1234
	putch('!');
}

unsigned char API_TestFlashThorough(void)
{
	unsigned int i,j;
	unsigned int TestFlash_ViewData;
	unsigned int test_flash_thorough_pattern_offset;
	unsigned int test_flash_thorough_total_errors;

	#define TEST_FLASH_THOROUGH_TOTAL_ERRORS_MAX 50
	
	//choose random number for offset
	test_flash_thorough_pattern_offset=(TMR0H<<8)+TMR0L;
//	EraseFlash();

	test_flash_thorough_total_errors=0;
	API_SetupFlash();

	for (j=0;j<=1;j++)
	{
		API_SelectFlashBank(j);
	
		for(i=0x1000; i < 0xffff; i++)
		{
			if (API_WriteFlash(i,(i+test_flash_thorough_pattern_offset))==FALSE)
			{
				putch('!');
				putinthex(i);
				test_flash_thorough_total_errors++;
				if (test_flash_thorough_total_errors>TEST_FLASH_THOROUGH_TOTAL_ERRORS_MAX)
				{
					putch('-');
					putinthex(test_flash_thorough_pattern_offset);
					putch('-');
					putchhex(j);
					putch('-');
					return FALSE;
				}
			}
		}

		b;

		for(i=0x1000; i < 0xffff; i++)
		{
			TestFlash_ViewData=API_ReadFlash(i);
			if (TestFlash_ViewData != (i+test_flash_thorough_pattern_offset))
			{
				putch('@');
				putinthex(i);
				putch('#');
				putinthex((i+test_flash_thorough_pattern_offset));
				putch('$');
				putinthex(TestFlash_ViewData);
				test_flash_thorough_total_errors++;
				if (test_flash_thorough_total_errors>TEST_FLASH_THOROUGH_TOTAL_ERRORS_MAX)
				{
					putch('-');
					putinthex(test_flash_thorough_pattern_offset);
					putch('-');
					return FALSE;
				}
			}
		}
	}
	b;	
	return TRUE;
}

